apiVersion: v1
kind: Service
metadata:
  name: {{ name }}
  namespace: {{ ns }}
  labels:
    app.kubernetes.io/name: {{ name }}
spec:
  selector:
    app.kubernetes.io/name: {{ name }}
  ports:
    - protocol: TCP
      port: {{ service_port | default(80) }}
      targetPort: {{ traffic_port }}
